import bz2
import logging

from construct import *

from simpleelf.elf_structs import ElfStructs
from simpleelf.elf_builder import ElfBuilder
from simpleelf import elf_consts

from iosfw.utils import get_elf_from_code

elf_structs = ElfStructs('>')

MAGIC = b'MZIP'

mzip_struct = Struct(
    'magic' / Const(MAGIC),
    
    Seek(0x38),
    'data_offset' / Hex(Int32ub),
    'entry' / Hex(Int32ub),
    'unknown0' / Int32ub,
    'compressed_size' / Hex(Int32ub),


    Seek(this.data_offset),
    'compressed_data' / Bytes(this.compressed_size),
    'comment' / GreedyBytes,
)

def unpack(fw, out):
    logging.debug('attempting MZIP packer')

    header_magic = MAGIC
    is_mzip = header_magic == fw.read(len(header_magic))
    fw.seek(0)

    if not is_mzip:
        logging.debug('not an MZIP file')
        return False

    logging.info('extracting MZIP packed file')

    mzip = mzip_struct.parse_stream(fw)

    e = get_elf_from_code(mzip.entry, bz2.decompress(mzip.compressed_data))
    out.write(e.build())

    logging.info('extracted successfully')
    return True



from construct import *

from simpleelf.elf_structs import ElfStructs
from simpleelf.elf_builder import ElfBuilder
from simpleelf import elf_consts

REGIONS_MAGIC = b'\xDE\xAD\x12\x34'

regions_struct = Struct(
    'magic' / Const(REGIONS_MAGIC),

    'unknown0' / Int32ub,
    'unknown1' / Int32ub,
    'unknown2' / Int32ub,
    'unknown3' / Int32ub,
    'unknown4' / Int32ub,

    'text' / Int32ub,
    'data' / Int32ub,
    'bss' / Int32ub,
    'heap' / Int32ub,

)

def guess_machine_type(code):
    # assuming cisco only uses mips and ppc cpus,
    # check for: isync; sync; should be enough to 
    # determine the cpu type
    if b'\x4C\x00\x01\x2C\x7C\x00\x04\xAC' in code:
        return elf_consts.EM_PPC
    else:
        return elf_consts.EM_MIPS

def get_elf_from_code(entry, main_image):
    e = ElfBuilder()
    e.set_endianity('>')

    regions_struct_offset = main_image.find(REGIONS_MAGIC)
    if -1 == regions_struct_offset:
        logging.warning("failed to locate regions magic. using single .text section")
        e.add_code_section('.text', entry, main_image)
    else:
        regions = regions_struct.parse(main_image[regions_struct_offset:
            regions_struct_offset+regions_struct.sizeof()])

        e.add_segment(entry, 
            elf_consts.PF_R | elf_consts.PF_W | elf_consts.PF_X, 
            len(main_image))

        e.add_code_section('.text', regions.text, main_image)
        e.add_empty_data_section('.data', regions.data, regions.bss - regions.data)
        e.add_empty_data_section('.bss', regions.bss, regions.heap - regions.bss)

    e.set_entry(entry)
    e.set_machine(guess_machine_type(main_image))

    return e



import logging
import os

import coloredlogs
import click

coloredlogs.install(level=logging.DEBUG)

from iosfw import mzip_packer
from iosfw import elf_packer

PACKERS = [elf_packer, mzip_packer]


@click.command()
@click.argument('fw', type=click.File('rb'))
@click.option('--out', type=click.File('wb'), default=None)
def unpack(fw, out=None):
    if out is None:
        out_filename = os.path.splitext(fw.name)[0] + '.elf'
        logging.info('out filename: {}'.format(out_filename))
        out = open(out_filename, 'wb')

    for p in PACKERS:
        if p.unpack(fw, out):
            return


@click.group()
def cli():
    pass


def main():
    cli.add_command(unpack)
    cli()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['iosfw']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.1,<8.0.0',
 'coloredlogs>=14.0,<15.0',
 'construct>=2.10.56,<3.0.0',
 'simpleelf>=0.1.2,<0.2.0']

entry_points = \
{'console_scripts': ['ios-fw = iosfw.cli:main']}

setup_kwargs = {
    'name': 'ios-fw',
    'version': '0.1.1',
    'description': 'IOS firmware utilities',
    'long_description': None,
    'author': 'Charlie Brown',
    'author_email': 'whysoyum@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
